



cd tests
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -c "import daal4py"
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -m unittest discover -v -p 'test*.py'
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest --pyargs ../daal4py/sklearn/
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest --pyargs ../onedal/
IF %ERRORLEVEL% NEQ 0 exit /B 1
python ../examples/daal4py/run_examples.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -m daal4py ../examples/daal4py/sycl/sklearn_sycl.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
